
--Copyright 2010 Gryc Ueusp

--This file is part of Collectible RPG Engine.

--Collectible RPG Engine is free software: you can redistribute it and/or modify
--it under the terms of the GNU General Public License as published by
--the Free Software Foundation, either version 3 of the License, or
--(at your option) any later version.

--Collectible RPG Engine is distributed in the hope that it will be useful,
--but WITHOUT ANY WARRANTY; without even the implied warranty of
--MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--GNU General Public License for more details.

--You should have received a copy of the GNU General Public License
--along with Collectible RPG Engine.  If not, see <http://www.gnu.org/licenses/>.


--Handles fighting monsters

require("dialog")
require("choice")
require("monster_formulas")

--[[

Draw: draw the monster sprites and health and monsters remaining bars.

Update: For things that need animating like health bars, 
    there should be a "current value" and a "display value". 
    When any of these differ, ignore any user input until we're done.
    We will skip animation until later.

Example: 
    Combat starts, things get initialized, monster sprites get drawn, etc.
    A dialog is shown "WILD X HAS APPEARED".
    After the dialog exits, a choice is shown, with combat type things.
    This choice calls combat sorts of functions, helper functions that 
        call functions from monster_formulas.lua.
    A monster takes damage. Their HP is set, but the display only does 
        another state when it's done animating.
]]

local messages = {}
local exit_okay = false
local num_monsters_participating = 1

local lastexp
local nextexp

local animating = false
local enemy_vis_hp = 0
local player_vis_hp = 0
local player_vis_exp = 0

function last_keymap()
	current_keymap = deepcopy(previous_keymaps[#previous_keymaps])
	table.remove(previous_keymaps)
end

function append_keymap()
	table.insert(previous_keymaps, deepcopy(current_keymap))
end

function combat_dialog(message, func, args)
	if func ~= nil then
		table.insert(messages, {message, func, args})
	else
		table.insert(messages, {message})
	end
end

function calc_bar_color(currenthp, maxhp)
	local green = {255/2, 255, 255/2}
    local yellow = {255, 255, 255/2}
    local red = {255, 255/2, 255/2}
    love.graphics.setColor(unpack(green))
    
	--HP Indicator Enemy
	if (currenthp/maxhp < 0.5) then
		love.graphics.setColor(unpack(yellow))
	end
	if (currenthp/maxhp < 0.25) then
		love.graphics.setColor(unpack(red))
	end
end

function do_move(user, target, move)
	print("Did move: "..move.name)
	
	local damage = move.effect(user, target, move)
	--dialog(string.format("%s used %s!", user.nickname, move.name))
	print(string.format("%s used %s!", user.nickname, move.name))
	combat_dialog(string.format("%s used %s!", user.nickname, move.name))
	print(string.format("%s dealt %d damage!", move.name, damage))
	combat_dialog(string.format("%s dealt %d damage!", move.name, damage))
	
	if damage > (target.currenthp/5) then
		--dialog("It was super effective!")
		combat_dialog("It was super effective!")
		print("It was super effective!")
	end
	
	print(damage, target.basestats.hp, target.currenthp)
	
	animating = true
end

function player_do_move(user, target, move)
	do_move(user, target, move)
	
	last_keymap()
	
	do_enemy_turn() --Turn end!
end

function do_enemy_turn()
	--This should be called after the user does something that would end their turn.
	
	if enemy.monsters[1].currenthp > 0 then
		--Item? Switch Out? Attack?
		
		--Heuristics first to determine what we can do and if it would be beneficial.
			--Not hurt? Prune all the healing items.
		--Go through several simulations and score each one. (Minimax)
		--Each iteration of the turns only has about 20 choices, but the number of choices is:
		--20, 400, 8000, 160k, 3.2mil, 64mil, 1.28bil, 25.6bil
		--To calculate two of the player's turns into the future, it would be 160 thousand possibilities.
		--Allow a maximum amount of calculation time before aborting to simulate intelligence (or lack thereof.)
		--We could, theoretically, prune the bottom 50% of "lines of thought" past a certain depth. "Remove the REALLY bad ideas."
		--TODO: Hide information from the AI, for now: "The Computer Is A Cheating Bastard" or "The All Seeing AI".
		
		--FIXME: AI just selects a move randomly.
		local move = enemy.monsters[1].moves[math.random(#enemy.monsters[1].moves)]
		--combat_dialog("Monster Attacked!", do_move, {enemy.monsters[1], player.monsters[1], move})
		do_move(enemy.monsters[1], player.monsters[1], move)
	end

	
	
end

function do_player_turn()
	--This should be called after the enemy goes.
	local combat_menu = {{"FIGHT", do_fight, {}}, 
						{"TEAM", print, {"Not Implemented"}}, 
						{"ITEMS", print, {"Not Implemented"}}, 
						{"RUN", print, {"Not Implemented"}}}
	choice_enter(combat_menu, combat_menu[1][1], 0, 0, "", "bottom")
end

function do_fight()
	print("Entering fight menu!")
	
	append_keymap()
	
	current_keymap = {}
	choice_map = {}
	
	for i=1,4 do
		if i <= #player.monsters[1].moves then
			local m = player.monsters[1].moves[i]
			table.insert(current_keymap, {''..i, m.name, do_move, player.monsters[1], enemy.monsters[1], m})
			table.insert(choice_map, {m.name, player_do_move, {player.monsters[1], enemy.monsters[1], m}})
		end
	end
	
	choice_enter(choice_map, choice_map[1][1], 50, 50, "", "bottom")
	
end



function combat_enter()
    --"Enters" the combat state
    --Sets up all the callbacks and such.
    
    table.insert(previous_callbacks, deepcopy(callbacks)) --Save the previous callbacks list
    callbacks = {}
    callbacks["exit"] = combat_exit
    callbacks["draw"] = combat_draw
    
    callbacks["keypressed"] = combat_keypressed
    callbacks["keyreleased"] = combat_keyreleased
    callbacks["update"] = combat_update
    
    combat_sprites = {}
    
    --Add the monster sprites to combat_sprites...or maybe just use goo
    --Set up animations via Goo to move monster sprites about
    --Also set up health bars, name text and messages. (goo.progressbar, goo.text, and goo.panel?)
    
    previous_backgroundcolor = {love.graphics.getBackgroundColor()}
    love.graphics.setBackgroundColor(255,255,255)
    frame = load_dialog_frame("/gfx/frames/fantasy_bold.png")
    
    icon = cached_load_image("gfx/capsule_fullsize.png")
    icongrey = cached_load_image("gfx/capsule_fullsize_greyscale.png")
    
    love.graphics.setFont(love.graphics.newFont(tilesize/2))
    
    current_message = ""
    
    if player.monsters[1].currenthp == nil or player.monsters[1].currenthp == 0 then
        player.monsters[1].currenthp = calc_hp(player.monsters[1])
    end
    if enemy.monsters[1].currenthp == nil or enemy.monsters[1].currenthp == 0 then
        enemy.monsters[1].currenthp = calc_hp(enemy.monsters[1])
    end
    
    exit_okay = false
    num_monsters_participating = 1
    
    do_player_turn()
    
    if #enemy.monsters > 1 then
		--Trainer
		--TODO: Iterate all monsters and assign nicknames
		if enemy.monsters[1].nickname ~= nil then
			enemy.monsters[1].nickname = "Enemy "..enemy.monsters[1].propername
		end
		dialog(string.format("Trainer sent out %s!", enemy.monsters[1].nickname))	
	else
		enemy.monsters[1].nickname = "Wild "..enemy.monsters[1].propername
		dialog(string.format("%s appeared!", enemy.monsters[1].nickname))	
	end
    
    previous_keymaps = {}
    current_keymap = {}
    
    player_vis_hp = player.monsters[1].currenthp
    player_vis_exp = player.monsters[1].currentexp
    enemy_vis_hp = enemy.monsters[1].currenthp
    
    combat_update() --Ensure that the appropriate vars get set before trying to draw.
    
end

function combat_exit()
    --The exit callback should NOT be called when entering another state, 
    --only when explicitly exiting and falling back to the previous state.
    
    love.graphics.setBackgroundColor(unpack(previous_backgroundcolor)) --Restore background color
    
    --Restore our callbacks when we're done.
    callbacks = deepcopy(previous_callbacks[#previous_callbacks])
    table.remove(previous_callbacks)
    print('exiting combat.')
end

function combat_init() --Should only be called once and from love.load()
    --Preload gui elements perhaps?
    
end

function combat_keypressed(key, unicode)
	
	for k,v in pairs(current_keymap) do
		if v[1] == key then
			--last_keymap()
			
			v[3](v[4], v[5], v[6])
			
			break
		end
	end

end

function combat_keyreleased(key, unicode)

end

function combat_update(dt)
    --Should do animations here.
    
    --Come back to the default menu.
    if(current_keymap == nil) then current_keymap = {} end
	if(#current_keymap == 0) then
	    current_keymap = {{"1","FIGHT", do_fight}, 
						  {"2","TEAM", print}, 
						  {"3","ITEMS", print},
						  {"4","RUN", print}}
	end
    
    current_message = ""
    breaklen = 30
    for k,v in pairs(current_keymap) do
		current_message = current_message .. v[1] .. ":" .. v[2]
		if k < #current_keymap then
			current_message = current_message .. "   "
		end
		if string.len(current_message) > breaklen then
			current_message = current_message .. "\n"
			breaklen = breaklen + 30
		end
    end
    --print(current_message)
    
    lastexp = player.monsters[1].exp_to_level(player.monsters[1].level)
    nextexp = (exp_to_level(player.monsters[1]) - lastexp)
    
    pmon = cached_load_image(player.monsters[1].sprites.combat.back)
    emon = cached_load_image(enemy.monsters[1].sprites.combat.front)
    
    if player.monsters[1].currenthp <= 0 and exit_okay == false then
		combat_dialog("Player lost the battle!")
		combat_dialog("This message is not shown but it's here to ensure the one before it is.")
		exit_okay = true
			
    end
    if enemy.monsters[1].currenthp <= 0 and exit_okay == false then
		combat_dialog(string.format("%s Fainted!", enemy.monsters[1].nickname))
		
		local exp_this_battle = calc_exp(enemy.monsters[1], player.monsters[1], num_monsters_participating)
		
		combat_dialog(string.format("%s got %d experience points!", player.monsters[1].nickname, exp_this_battle))
		
		--TODO: Iterate through participating monsters and give exp as normal.
		local m = player.monsters[1]
		local hpratio = m.currenthp/calc_hp(m) --Store the hp ratio so the monster doesnt "lose" HP on level up.
		if m.currentexp + exp_this_battle > exp_to_level(m) then
			m.level = m.level + 1
			combat_dialog(string.format("%s levelled up to %d!", m.nickname, m.level))
			m.currenthp = calc_hp(m)*hpratio
		end
		
		player.monsters[1].currentexp = player.monsters[1].currentexp + exp_this_battle
		
		combat_dialog("This message is not shown but it's here to ensure the one before it is.")
		
		exit_okay = true
    end
    
    if animating == false then
		if choice_active == false then
			if #messages == 1 and exit_okay == false then
				do_player_turn()
			end
			for k,v in pairs(deepcopy(messages)) do
				dialog(v[1])
				table.remove(messages, k)
				break
			end
		end	
		if exit_okay == true and #messages == 0 then
			combat_exit()
		end
	elseif animating == true then
		--Do our animation.
		
		--HP?
		if round(enemy_vis_hp) ~= enemy.monsters[1].currenthp then
			if enemy_vis_hp > enemy.monsters[1].currenthp then
				enemy_vis_hp = enemy_vis_hp - dt*5
			elseif enemy_vis_hp < enemy.monsters[1].currenthp then
				enemy_vis_hp = enemy_vis_hp + dt*5
			end
		elseif round(player_vis_hp) ~= player.monsters[1].currenthp then
			if player_vis_hp > player.monsters[1].currenthp then
				player_vis_hp = player_vis_hp - dt*5
			elseif player_vis_hp < player.monsters[1].currenthp then
				player_vis_hp = player_vis_hp + dt*5
			end
        elseif round(player_vis_exp) ~= player.monsters[1].currentexp then
            if player_vis_exp < player.monsters[1].currentexp then
                player_vis_exp = player_vis_exp + dt*5
            end
		elseif player_attacking == true or enemy_attacking == true then
			--Attack?
		else
			animating = false
		end
		
	end
end

function combat_draw()

	--Background
	love.graphics.rectangle("fill", 0, 0,  windowwidth, windowheight)

    --monster sprites
    love.graphics.setColor(255, 255, 255)
    
    love.graphics.draw(pmon, 0, windowheight-(pmon:getHeight()/2)-(windowheight-(tilesize*7.5)), 0, 320/pmon:getWidth(), 320/pmon:getWidth())
    love.graphics.draw(emon, windowwidth-(emon:getWidth()/2), 0, 0, 320/emon:getWidth(), 320/emon:getWidth())
    
    --frames for our text and other ui elements
    
    --monster names
    love.graphics.setColor(0, 0, 0)
    love.graphics.print(string.format("L%d %s", enemy.monsters[1].level, enemy.monsters[1].nickname), windowwidth/50, windowheight/100)
    love.graphics.print(string.format("L%d %s", player.monsters[1].level, player.monsters[1].nickname), windowwidth/1.8, windowheight/2.1)
    
    --hp bars
    local p_currenthp, p_maxhp = player.monsters[1].currenthp, calc_hp(player.monsters[1])
    local e_currenthp, e_maxhp = enemy.monsters[1].currenthp, calc_hp(enemy.monsters[1])
    --print(p_maxhp, p_currenthp, e_maxhp, e_currenthp)
    
    local bar_width, bar_height  = windowwidth/2.5, windowheight/40
    
    --HP Indicator
	--HP Indicator Enemy
	calc_bar_color(e_currenthp, e_maxhp)
    love.graphics.rectangle("fill", windowwidth/50, (windowheight/20)+15, bar_width*(enemy_vis_hp/e_maxhp), bar_height)
	
	--HP Indicator Player
	calc_bar_color(p_currenthp, p_maxhp)
    love.graphics.rectangle("fill", windowwidth/1.8, (windowheight/1.9)+15,  bar_width*(player_vis_hp/p_maxhp), bar_height)
    --TODO: Text representation of player's monster's HP.
    love.graphics.setColor(0, 0, 0)
    love.graphics.print(string.format("HP: %d/%d", p_currenthp, p_maxhp), windowwidth/1.8, windowheight/1.55)
    
    --HP Outline
    love.graphics.setColor(0, 255/2, 0)
    love.graphics.rectangle("line", windowwidth/50, (windowheight/20)+15, bar_width, bar_height+5)
    love.graphics.rectangle("line", windowwidth/1.8, (windowheight/1.9)+15, bar_width, bar_height+5)
    
    --monster count for enemy OR caught this monster yet indicator
    love.graphics.setColor(255, 255, 255)
    local wildiconposx = ((windowwidth/50)+bar_width)-20
    local wildiconposy = (windowheight/20)-20
    if #enemy.monsters > 1 then
		--Trainer
		for m in enemy.monsters do
			if m.currenthp == 0 then
				--Empty capsule
			else
				--Full Capsule
			end
		end
	else
		--Wild
		if player.monsterindex[enemy.monsters[1].propername] ~= nil then
			--Already caught
			love.graphics.draw(icon, wildiconposx, wildiconposy, 0, .025, .025, 0, 0)
		else
			--Not caught yet!
			love.graphics.draw(icongrey, wildiconposx, wildiconposy, 0, .025, .025, 0, 0)
		end
	end
    
    --xp bar for your monster
    love.graphics.setColor(255/2, 255/2, 255)
    
    love.graphics.rectangle("fill", windowwidth/1.8, (windowheight/1.9)+55,  bar_width*((player_vis_exp-lastexp)/nextexp), bar_height)
    
    love.graphics.setColor(0, 0, 255)
    love.graphics.rectangle("line", windowwidth/1.8, (windowheight/1.9)+55, bar_width, bar_height)
    
    love.graphics.setColor(255, 255, 255)
    
    --dialog frame
	local draw_retval = {draw_dialog_frame(frame, 0, tilesize*7.5, windowwidth, windowheight-(tilesize*7.5))}
	--local corner_height, corner_width, corner_sy, corner_sx, dialog_x, dialog_y, dialog_w, dialog_h, horizscale, vertscale = unpack(draw_retval)
    
    --dialog text
    --draw_dialog_text(current_message, corner_width, corner_height, corner_sx, corner_sy, dialog_y)
    

end
