
--Copyright 2010 Gryc Ueusp

--This file is part of Collectible RPG Engine.

--Collectible RPG Engine is free software: you can redistribute it and/or modify
--it under the terms of the GNU General Public License as published by
--the Free Software Foundation, either version 3 of the License, or
--(at your option) any later version.

--Collectible RPG Engine is distributed in the hope that it will be useful,
--but WITHOUT ANY WARRANTY; without even the implied warranty of
--MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
--GNU General Public License for more details.

--You should have received a copy of the GNU General Public License
--along with Collectible RPG Engine.  If not, see <http://www.gnu.org/licenses/>.



require "lxp"

print_table = table_print --An alias to save my sanity.
require("tablesave")

function _load_tiled_map(filedata)
    --- Load a map from the Tiled program.
    -- filedata Should be a table of lines.
    -- Returns a table containing a table per Layer, and each table should be rows with column members 
    --      inside so we can index them directly by x and y so we dont draw things we dont need to.
    -- Also include the scale multiplier in each direction so we dont have to call getWidth/getHeight every frame.
    -- The Images are cached for lower memory usage and speed.
    
    --If a map layer's name has "(not rendered)" in it (case insensitive), 
    --  or if the map layer's name has "physical" or "events" in it (case insensitive),
    --  or if it has a property "dontdraw" set to "true" 
    --  or if the layer's xml element has a "visible" property set to "0"
    --Then, set the map layer's visible property to false.
    --Else, dont even set one.
        
    layers = {}
    layerorder = {}
    local tilesets = {}
    local tiles = {}
    
    local mapheight, mapwidth = nil, nil
    
    local lastlayer = nil
    local lasttile = nil 
    local lastx, lasty = 1,1
    local lastheight, lastwidth = 0, 0
    local prevfirstgid = 0
    
    local default = load_from_spritesheet_by_number("/gfx/tileset.png", 32, 32, 1)
    
    local callstack = {}
    local callbackstack = {}
    
    lxpcallbacks = {
    StartElement = function (parser, name, attributes)
        table.insert(callstack, name)
        table.insert(callbackstack, lxpcallbacks.StartElement)
        
        --Load any map attributes we need from the toplevel <map>
        if name == "map" and #callstack == 1 then
            mapheight, mapwidth = attributes.height, attributes.width
        end
        
        --Load any map properties we need from the toplevel <map>
        if name == "property" and #callstack == 3 then
            --Dont need any properties at the moment.
        end
        
        --Any attributes we need from this <tileset>?
        if name == "tileset" and #callstack == 2 then
            if attributes.source ~= nil then
                print("WARNING: External tilesets are not supported yet, please import them before saving your map.")
            end
            
            for x=prevfirstgid,attributes.firstgid do
                tiles[x] = {}
            end
            tilesets[attributes.firstgid] = {}
            prevfirstgid = attributes.firstgid
        end
        
        --Now does that tileset have any images we're interested in?
        if name == "image" and #callstack == 3 then
            tilesets[prevfirstgid] = attributes.source
        end
        
        --Lets load any map>tileset>tiles that have attributes.
        if name == "tile" and #callstack == 3 then
            lasttile = tostring(tonumber(prevfirstgid)+tonumber(attributes.id))
            tiles[lasttile] = {}
        end
        
        --Get all map>tileset>tiles' properties.
        if name == "property" and #callstack == 5 then
            tiles[lasttile][attributes.name] = attributes.value
        end
        
        --Get all map>layers and their info
        if name == "layer" and #callstack == 2 then
            -- Make space for all the tiles that will go in here so we dont have to later.
            print("Loading layer "..attributes.name.."...")
            layerorder[#layerorder+1] = attributes.name
            lastx, lasty = 1,1
            layers[attributes.name] = {}
            lastheight = attributes.height
            lastwidth = attributes.width
            for x=1,attributes.width do
                layers[attributes.name][x] = {}
                for y=1,attributes.height do
                    layers[attributes.name][x][y] = {}
                end
            end
            
            local name = attributes.name:lower()
            if string.find("(not rendered)", name) or string.find("events", name) or string.find("physical", name) or attributes.visible == "0" then
                layers[attributes.name].visible = false
            else
                layers[attributes.name].visible = true
            end
            
            lastlayer = attributes.name
        end
        
        --All map>layer>data>tile info.
        if name == "tile" and #callstack == 4 then
            --{image=nil, mult=nil, physical=false, onStep="print('1,1')", onActivate=nil}
            local data = {image=nil, mult=nil, physical=false, onStep=nil, onActivate=nil, onInit=nil}
            
            
            --For keys in tilesets,
            --If the tile we're looking up's gid is greater than the previous one, but less than the current one, 
            --The tileset we want is the previous one.
            --If none match, it's the last in the line.
            local srcimg, previous
            previous = 1
            for k, v in pairs(tilesets) do
                if tonumber(attributes.gid) >= tonumber(previous) then
                    if tonumber(attributes.gid) <= tonumber(k) then
                        srcimg = previous
                    end
                end
                previous = k
            end
            if srcimg == nil then
                srcimg = previous
            end
            
            local phys, onstep, onact
            local tilenum = tonumber(attributes.gid)-srcimg
            local gid = tostring(tonumber(attributes.gid))
            if tiles[gid] ~= nil then
                if tiles[gid].physical == "true" then
                    phys = true
                else
                    phys = false
                end
                onstep = tiles[gid].onStep
                onact = tiles[gid].onActivate
                oninit = tiles[gid].onInit
            else
                phys = false
                onstep = nil
                onact = nil
                oninit = nil
            end
            local tset = tilesets[tostring(tonumber(srcimg))]
            
            local imgpath = string.gsub(tset, "../../../", "")
            
            data.image = load_from_spritesheet_by_number(imgpath, 32, 32, tilenum)
            data.mult = 2
            data.physical = phys
            data.onStep = onstep
            data.onActivate = onact
            data.onInit = oninit
            layers[lastlayer][lastx][lasty] = data --Data goes here
            
            --Update our coords.
            if lastx == tonumber(lastwidth) then
                lastx = 1
                lasty = lasty + 1
            else 
                lastx = lastx + 1
            end
            if lasty > tonumber(lastheight) and lastx > 1 then
                --We expect it to go exactly one over in the count, any more and it's definitely an error.
                print("WARNING: Too many tiles in map for map dimensions! They will not appear in the rendered map.", lastx, lasty)
            end
        end
        
        --map>layer>properties>property, or layer's properties.
        if name == "property" and #callstack == 4 then
            if attributes.dontdraw ~= nil then --If dont draw is set, make the layer invisible.
                layers[lastlayer].visible = false
            end
        end
        
    end,
    
    EndElement = function (parser, name)
        lxpcallbacks.StartElement = callbackstack[#callbackstack]
        table.remove(callbackstack)
--~         local bcrumb = ""
--~         for x=1,#callstack do
--~             bcrumb = bcrumb..">"..callstack[x]
--~         end
--~         print(bcrumb)
        table.remove(callstack)
    end

    }
    
    p = lxp.new(lxpcallbacks)
    for k,l in pairs(filedata) do  -- iterate lines
        p:parse(l)          -- parses the line
        p:parse("\n")       -- parses the end of line
    end
    p:parse()               -- finishes the document
    p:close()               -- closes the parser


    return layers, layerorder

end

function load_tiled_map(filename)
    --- Load a tiled map via LOVE
    
    --If there's a .tbl file for this map, just load and return that instead.
    --TODO: Check if the .tbl is older than the .tmx. Load whichever is newer, or solely the .tbl on windows.
    if love.filesystem.isFile(filename..".tbl") then
        print("Loading cached map "..filename..".tbl")
        local t, err = table.load(love.filesystem.read(filename..".tbl").."--|")
        return t[1], t[2]
    end
    
    if love.filesystem.isFile(filename) then
        print("Loading map "..filename)
        
        local lines = {}
        for l in love.filesystem.lines(filename) do
            table.insert(lines, l)
        end
        
        return _load_tiled_map(lines)
    end
end

function load_tiled_map_nonlove(filename)
    --- Load a tiled map via the standard Lua IO library
    -- Use this one if you're not using LOVE.
    
    file = io.open (filename, "r")
    if file then
        print("Loading map "..filename)
        
        local lines = {}
        for l in io.lines(filename) do
            table.insert(lines, l)
        end
        io.close(file)
        
        return _load_tiled_map(lines)
    end
    
end
